import json
import os
from dataclasses import dataclass, field
from typing import Any, Dict, List

import requests
from requests.exceptions import HTTPError

from ..bridge import BridgeUsername
from ..error import SlackException
from ..github.models.pull_request import PullRequest
from ..github.models.user import User


@dataclass(frozen=True)
class Client:
    usernames: List[BridgeUsername]
    icon: bool
    exclude_users: [str]
    _webhook_url: str = field(init=False, default_factory=lambda: os.environ["SLACK_URL"])

    def _convert_github_to_slack(self, user: User) -> str:
        for username in self.usernames:
            if username.github == user.login:
                return username.slack
        else:
            return user.login

    def _get_section(self, pull: PullRequest) -> List[Dict[str, Any]]:
        # title section
        title_section: Dict[str, Any] = {
            "type": "section",
            "text": {
                "type": "mrkdwn",
                "text": f"[#{pull.number}] <{pull.html_url}|{pull.title}> (_{pull.user.login}_)",
            },
            "accessory": {
                "type": "button",
                "text": {
                    "type": "plain_text",
                    "text": "Review",
                },
                "url": pull.html_url,
            },
        }

        # reviewer section
        reviewer_section: Dict[str, Any] = {"type": "context", "elements": []}
        requested_reviewers = list(filter(lambda obj: obj.login not in self.exclude_users, pull.requested_reviewers))
        if requested_reviewers:
            reviewer_section["elements"] += [
                {
                    "type": "plain_text",
                    "text": "Waiting on",
                },
            ]
            if self.icon:
                for reviewer in requested_reviewers:
                    reviewer_section["elements"] += [
                        {
                            "type": "image",
                            "image_url": reviewer.avatar_url,
                            "alt_text": reviewer.login,
                        },
                        {
                            "type": "mrkdwn",
                            "text": self._convert_github_to_slack(user=reviewer),
                        },
                    ]
            else:
                reviewer_section["elements"] += [
                    {
                        "type": "mrkdwn",
                        "text": " ".join(map(self._convert_github_to_slack, requested_reviewers)),
                    }
                ]
        else:
            reviewer_section["elements"] += [{"type": "plain_text", "text": "Waiting review by anyone."}]

        return [
            title_section,
            reviewer_section,
        ]

    def _build_block(self, repo: str, pulls: List[PullRequest], total_pulls: int) -> List[Dict[str, Any]]:
        blocks = [
            {
                "type": "section",
                "text": {
                    "type": "mrkdwn",
                    "text": f"*Pending review on _<https://github.com/{repo}|{repo}>_*",
                },
            },
        ]

        for section in map(lambda pull: self._get_section(pull=pull), pulls):
            blocks += section

        if len(pulls) < total_pulls:
            blocks += [
                {
                    "type": "section",
                    "text": {
                        "type": "mrkdwn",
                        "text": f"Number of remaining: *{total_pulls - len(pulls)}*",
                    },
                    "accessory": {
                        "type": "button",
                        "text": {
                            "type": "plain_text",
                            "text": "Watch",
                        },
                        "url": f"https://github.com/{repo}/pulls",
                    },
                },
            ]

        blocks += [
            {
                "type": "divider",
            },
        ]

        return blocks

    def post(self, repo: str, pulls: List[PullRequest], total_pulls: int) -> None:
        if not pulls:
            return

        payload: Dict[str, Any] = {
            "text": f"Waiting your review on {repo}.",
            "blocks": self._build_block(repo=repo, pulls=pulls, total_pulls=total_pulls),
        }

        res = requests.post(url=self._webhook_url, data=json.dumps(payload))
        try:
            res.raise_for_status()
        except HTTPError:
            raise SlackException(status_code=res.status_code, content=res.content.decode("utf-8"))



from pydantic import BaseModel, Field


class User(BaseModel):
    login: str
    user_id: int = Field(alias="id")
    user_type: str = Field(alias="type")
    avatar_url: str

    def __str__(self) -> str:
        user = f"@{self.login}"
        if self.user_type != "User":
            user += f" [{self.user_type}]"
        return user



from typing import List, Optional

from pydantic import BaseModel, Field

from .user import User


class PullRequest(BaseModel):
    pr_id: int = Field(alias="id")
    node_id: str
    html_url: str
    number: int
    state: str
    locked: bool
    title: str
    user: User
    assignees: List[User] = Field(default_factory=lambda: list)
    requested_reviewers: List[User] = Field(default_factory=lambda: list)
    draft: Optional[bool]

    def __str__(self) -> str:
        title = f"Pull Request [#{self.number}] {self.title} from {str(self.user)}"
        if self.draft:
            title += "in draft"
        return title



import json
import os
import sys
from pathlib import Path
from typing import List, Optional

import click

from . import __version__
from .bridge import BridgeUsername
from .error import SlackException
from .github import api as github_api
from .slack import api as slack_api


def get_bridge_usernames(file_username: Optional[str]) -> List[BridgeUsername]:
    if file_username is None:
        return []
    return list(
        map(
            lambda item: BridgeUsername(**item),
            json.loads(Path(file_username).read_text()),
        )
    )


@click.version_option(version=__version__)
@click.command(
    help="""GitHub repository's pull requests slack notifier\n
Required environments variables\n
    - GITHUB_TOKEN: Your GitHub Personal Access Token.\n
    - SLACK_URL: Your Slack webhook url."""
)
@click.option(
    "--repo",
    "-r",
    type=str,
    help="GitHub repository name (owner/repo).",
    required=True,
)
@click.option(
    "--file-username",
    "-u",
    type=str,
    help="Corresponding files for GitHub and Slack usernames. (see. example in README.md)",
)
@click.option(
    "--limit",
    "-l",
    type=int,
    default=20,
    help="Number of Pull Requests to notify Slack.",
    required=True,
)
@click.option(
    "--icon",
    "-i",
    is_flag=True,
    type=bool,
    default=False,
    help="Give GitHub icons to Slack notifications.",
    required=False,
)
@click.option(
    "--exclude-users",
    "-eu",
    type=str,
    default=[],
    multiple=True,
    help="GitHub users to remove from reviewers upon notification.",
    required=False,
)
def main(repo: str, file_username: Optional[str], limit: int, icon: bool, exclude_users: [str]) -> None:
    if limit > 20:
        raise ValueError("Cannot set more than 20 items.")

    gh = github_api.Client(github_token=os.environ["GITHUB_TOKEN"])
    # get pull requests
    pulls = gh.get_pulls(repo=repo, limit=limit)
    # get total pull requests count
    total_pulls = gh.get_total_pulls(repo=repo)
    # get GitHub and GitHub bridge username
    usernames = get_bridge_usernames(file_username=file_username)

    # send slack
    sl = slack_api.Client(usernames=usernames, icon=icon, exclude_users=exclude_users)
    try:
        sl.post(repo=repo, pulls=pulls, total_pulls=total_pulls)
    except SlackException as e:
        print(f"Slack Exception {e.response['detail']}", end=" ")
        sys.exit(1)


if __name__ == "__main__":
    main()
